@@ -19,10 +19,11 @@
 #include "filehash.h"
 #include "verify.h"
 #include "profile.h"
+#include "server.h"
 
 #include "lib/getopt.h"
 
-static char fio_version_string[] = "fio 1.59";
+const char fio_version_string[] = FIO_VERSION;
 
 #define FIO_RANDSEED		(0xb1899bedUL)
 
@@ -44,7 +45,10 @@ char **job_sections = NULL;
 int nr_job_sections = 0;
 char *exec_profile = NULL;
 int warnings_fatal = 0;
-int terse_version = 2;
+int terse_version = 3;
+int is_backend = 0;
+int nr_clients = 0;
+int log_syslog = 0;
 
 int write_bw_log = 0;
 int read_only = 0;
@@ -58,6 +62,29 @@ unsigned int fio_debug_jobno = -1;
 unsigned int *fio_debug_jobp = NULL;
 
 static char cmd_optstr[256];
+static int did_arg;
+
+const fio_fp64_t def_percentile_list[FIO_IO_U_LIST_MAX_LEN] = {
+	{ .u.f	=  1.00 },
+	{ .u.f	=  5.00 },
+	{ .u.f	= 10.00 },
+	{ .u.f	= 20.00 },
+	{ .u.f	= 30.00 },
+	{ .u.f	= 40.00 },
+	{ .u.f	= 50.00 },
+	{ .u.f	= 60.00 },
+	{ .u.f	= 70.00 },
+	{ .u.f	= 80.00 },
+	{ .u.f	= 90.00 },
+	{ .u.f	= 95.00 },
+	{ .u.f	= 99.00 },
+	{ .u.f	= 99.50 },
+	{ .u.f	= 99.90 },
+	{ .u.f	= 99.95 },
+	{ .u.f	= 99.99 },
+};
+
+#define FIO_CLIENT_FLAG		(1 << 16)
 
 /*
  * Command line options. These will contain the above, plus a few
@@ -67,117 +94,203 @@ static struct option l_opts[FIO_NR_OPTIONS] = {
 	{
 		.name		= (char *) "output",
 		.has_arg	= required_argument,
-		.val		= 'o',
+		.val		= 'o' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "timeout",
 		.has_arg	= required_argument,
-		.val		= 't',
+		.val		= 't' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "latency-log",
 		.has_arg	= required_argument,
-		.val		= 'l',
+		.val		= 'l' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "bandwidth-log",
 		.has_arg	= required_argument,
-		.val		= 'b',
+		.val		= 'b' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "minimal",
 		.has_arg	= optional_argument,
-		.val		= 'm',
+		.val		= 'm' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "version",
 		.has_arg	= no_argument,
-		.val		= 'v',
+		.val		= 'v' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "help",
 		.has_arg	= no_argument,
-		.val		= 'h',
+		.val		= 'h' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "cmdhelp",
 		.has_arg	= optional_argument,
-		.val		= 'c',
+		.val		= 'c' | FIO_CLIENT_FLAG,
+	},
+	{
+		.name		   = (char *) "enghelp",
+		.has_arg	= optional_argument,
+		.val		    = 'i' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "showcmd",
 		.has_arg	= no_argument,
-		.val		= 's',
+		.val		= 's' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "readonly",
 		.has_arg	= no_argument,
-		.val		= 'r',
+		.val		= 'r' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "eta",
 		.has_arg	= required_argument,
-		.val		= 'e',
+		.val		= 'e' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "debug",
 		.has_arg	= required_argument,
-		.val		= 'd',
+		.val		= 'd' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "section",
 		.has_arg	= required_argument,
-		.val		= 'x',
+		.val		= 'x' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "alloc-size",
 		.has_arg	= required_argument,
-		.val		= 'a',
+		.val		= 'a' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "profile",
 		.has_arg	= required_argument,
-		.val		= 'p',
+		.val		= 'p' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "warnings-fatal",
 		.has_arg	= no_argument,
-		.val		= 'w',
+		.val		= 'w' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "max-jobs",
 		.has_arg	= required_argument,
-		.val		= 'j',
+		.val		= 'j' | FIO_CLIENT_FLAG,
 	},
 	{
 		.name		= (char *) "terse-version",
 		.has_arg	= required_argument,
-		.val		= 'V',
+		.val		= 'V' | FIO_CLIENT_FLAG,
+	},
+	{
+		.name		= (char *) "server",
+		.has_arg	= optional_argument,
+		.val		= 'S',
+	},
+	{	.name		= (char *) "daemonize",
+		.has_arg	= required_argument,
+		.val		= 'D',
+	},
+	{
+		.name		= (char *) "client",
+		.has_arg	= required_argument,
+		.val		= 'C',
 	},
 	{
 		.name		= NULL,
 	},
 };
 
-FILE *get_f_out()
+static void free_shm(void)
 {
-	return f_out;
+	struct shmid_ds sbuf;
+
+	if (threads) {
+		void *tp = threads;
+
+		threads = NULL;
+		file_hash_exit();
+		flow_exit();
+		fio_debug_jobp = NULL;
+		shmdt(tp);
+		shmctl(shm_id, IPC_RMID, &sbuf);
+	}
+
+	scleanup();
 }
 
-FILE *get_f_err()
+/*
+ * The thread area is shared between the main process and the job
+ * threads/processes. So setup a shared memory segment that will hold
+ * all the job info. We use the end of the region for keeping track of
+ * open files across jobs, for file sharing.
+ */
+static int setup_thread_area(void)
 {
-	return f_err;
+	void *hash;
+
+	if (threads)
+		return 0;
+
+	/*
+	 * 1024 is too much on some machines, scale max_jobs if
+	 * we get a failure that looks like too large a shm segment
+	 */
+	do {
+		size_t size = max_jobs * sizeof(struct thread_data);
+
+		size += file_hash_size;
+		size += sizeof(unsigned int);
+
+		shm_id = shmget(0, size, IPC_CREAT | 0600);
+		if (shm_id != -1)
+			break;
+		if (errno != EINVAL && errno != ENOMEM) {
+			perror("shmget");
+			break;
+		}
+
+		max_jobs >>= 1;
+	} while (max_jobs);
+
+	if (shm_id == -1)
+		return 1;
+
+	threads = shmat(shm_id, NULL, 0);
+	if (threads == (void *) -1) {
+		perror("shmat");
+		return 1;
+	}
+
+	memset(threads, 0, max_jobs * sizeof(struct thread_data));
+	hash = (void *) threads + max_jobs * sizeof(struct thread_data);
+	fio_debug_jobp = (void *) hash + file_hash_size;
+	*fio_debug_jobp = -1;
+	file_hash_init(hash);
+
+	flow_init();
+
+	return 0;
 }
 
 /*
  * Return a free job structure.
  */
-static struct thread_data *get_new_job(int global, struct thread_data *parent)
+static struct thread_data *get_new_job(int global, struct thread_data *parent,
+				       int preserve_eo)
 {
 	struct thread_data *td;
 
 	if (global)
 		return &def_thread;
+	if (setup_thread_area()) {
+		log_err("error: failed to setup shm segment\n");
+		return NULL;
+	}
 	if (thread_number >= max_jobs) {
 		log_err("error: maximum number of jobs (%d) reached.\n",
 				max_jobs);
@@ -187,10 +300,14 @@ static struct thread_data *get_new_job(int global, struct thread_data *parent)
 	td = &threads[thread_number++];
 	*td = *parent;
 
+	td->io_ops = NULL;
+	if (!preserve_eo)
+		td->eo = NULL;
+
 	td->o.uid = td->o.gid = -1U;
 
 	dup_files(td, parent);
-	options_mem_dupe(td);
+	fio_options_mem_dupe(td);
 
 	profile_add_hooks(td);
 
@@ -204,10 +321,15 @@ static void put_job(struct thread_data *td)
 		return;
 
 	profile_td_exit(td);
+	flow_exit_job(td);
 
 	if (td->error)
 		log_info("fio: %s\n", td->verror);
 
+	fio_options_free(td);
+	if (td->io_ops)
+		free_ioengine(td);
+
 	memset(&threads[td->thread_number - 1], 0, sizeof(*td));
 	thread_number--;
 }
@@ -215,21 +337,19 @@ static void put_job(struct thread_data *td)
 static int __setup_rate(struct thread_data *td, enum fio_ddir ddir)
 {
 	unsigned int bs = td->o.min_bs[ddir];
-	unsigned long long bytes_per_sec;
 
 	assert(ddir_rw(ddir));
 
 	if (td->o.rate[ddir])
-		bytes_per_sec = td->o.rate[ddir];
+		td->rate_bps[ddir] = td->o.rate[ddir];
 	else
-		bytes_per_sec = td->o.rate_iops[ddir] * bs;
+		td->rate_bps[ddir] = td->o.rate_iops[ddir] * bs;
 
-	if (!bytes_per_sec) {
+	if (!td->rate_bps[ddir]) {
 		log_err("rate lower than supported\n");
 		return -1;
 	}
 
-	td->rate_nsec_cycle[ddir] = 1000000000ULL / bytes_per_sec;
 	td->rate_pending_usleep[ddir] = 0;
 	return 0;
 }
@@ -280,11 +400,18 @@ static int fixup_options(struct thread_data *td)
 	}
 
 	/*
-	 * only really works for sequential io for now, and with 1 file
+	 * only really works with 1 file
 	 */
-	if (o->zone_size && td_random(td) && o->open_files == 1)
+	if (o->zone_size && o->open_files == 1)
 		o->zone_size = 0;
 
+	/*
+	 * If zone_range isn't specified, backward compatibility dictates it
+	 * should be made equal to zone_size.
+	 */
+	if (o->zone_size && !o->zone_range)
+		o->zone_range = o->zone_size;
+
 	/*
 	 * Reads can do overwrites, we always need to pre-create the file
 	 */
@@ -421,6 +548,28 @@ static int fixup_options(struct thread_data *td)
 	}
 #endif
 
+#ifdef WIN32
+	/*
+	 * Windows doesn't support O_DIRECT or O_SYNC with the _open interface,
+	 * so fail if we're passed those flags
+	 */
+	if ((td->io_ops->flags & FIO_SYNCIO) && (td->o.odirect || td->o.sync_io)) {
+		log_err("fio: Windows does not support direct or non-buffered io with"
+				" the synchronous ioengines. Use the 'windowsaio' ioengine"
+				" with 'direct=1' and 'iodepth=1' instead.\n");
+		ret = 1;
+	}
+#endif
+
+	/*
+	 * For fully compressible data, just zero them at init time.
+	 * It's faster than repeatedly filling it.
+	 */
+	if (td->o.compress_percentage == 100) {
+		td->o.zero_buffers = 1;
+		td->o.compress_percentage = 0;
+	}
+
 	return ret;
 }
 
@@ -526,29 +675,60 @@ void td_fill_rand_seeds(struct thread_data *td)
 	init_rand_seed(&td->buf_state, td->rand_seeds[7]);
 }
 
+
 /*
- * Initialize the various random states we need (random io, block size ranges,
- * read/write mix, etc).
+ * Initializes the ioengine configured for a job, if it has not been done so
+ * already.
  */
-static int init_random_state(struct thread_data *td)
+int ioengine_load(struct thread_data *td)
 {
-	int fd;
+	const char *engine;
 
-	fd = open("/dev/urandom", O_RDONLY);
-	if (fd == -1) {
-		td_verror(td, errno, "open");
+	/*
+	 * Engine has already been loaded.
+	 */
+	if (td->io_ops)
+		return 0;
+
+	engine = get_engine_name(td->o.ioengine);
+	td->io_ops = load_ioengine(td, engine);
+	if (!td->io_ops) {
+		log_err("fio: failed to load engine %s\n", engine);
 		return 1;
 	}
 
-	if (read(fd, td->rand_seeds, sizeof(td->rand_seeds)) <
-	    (int) sizeof(td->rand_seeds)) {
-		td_verror(td, EIO, "read");
-		close(fd);
-		return 1;
+	if (td->io_ops->option_struct_size && td->io_ops->options) {
+		/*
+		 * In cases where td->eo is set, clone it for a child thread.
+		 * This requires that the parent thread has the same ioengine,
+		 * but that requirement must be enforced by the code which
+		 * cloned the thread.
+		 */
+		void *origeo = td->eo;
+		/*
+		 * Otherwise use the default thread options.
+		 */
+		if (!origeo && td != &def_thread && def_thread.eo &&
+		    def_thread.io_ops->options == td->io_ops->options)
+			origeo = def_thread.eo;
+
+		options_init(td->io_ops->options);
+		td->eo = malloc(td->io_ops->option_struct_size);
+		/*
+		 * Use the default thread as an option template if this uses the
+		 * same options structure and there are non-default options
+		 * used.
+		 */
+		if (origeo) {
+			memcpy(td->eo, origeo, td->io_ops->option_struct_size);
+			options_mem_dupe(td->eo, td->io_ops->options);
+		} else {
+			memset(td->eo, 0, td->io_ops->option_struct_size);
+			fill_default_options(td->eo, td->io_ops->options);
+		}
+		*(struct thread_data **)td->eo = td;
 	}
 
-	close(fd);
-	td_fill_rand_seeds(td);
 	return 0;
 }
 
@@ -562,7 +742,6 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	const char *ddir_str[] = { NULL, "read", "write", "rw", NULL,
 				   "randread", "randwrite", "randrw" };
 	unsigned int i;
-	const char *engine;
 	char fname[PATH_MAX];
 	int numjobs, file_alloced;
 
@@ -583,12 +762,8 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	if (profile_td_init(td))
 		goto err;
 
-	engine = get_engine_name(td->o.ioengine);
-	td->io_ops = load_ioengine(td, engine);
-	if (!td->io_ops) {
-		log_err("fio: failed to load engine %s\n", engine);
+	if (ioengine_load(td))
 		goto err;
-	}
 
 	if (td->o.use_thread)
 		nr_thread++;
@@ -616,6 +791,15 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	if (fixup_options(td))
 		goto err;
 
+	flow_init_job(td);
+
+	/*
+	 * IO engines only need this for option callbacks, and the address may
+	 * change in subprocesses.
+	 */
+	if (td->eo)
+		*(struct thread_data **)td->eo = NULL;
+
 	if (td->io_ops->flags & FIO_DISKLESSIO) {
 		struct fio_file *f;
 
@@ -627,9 +811,9 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 
 	td->ts.clat_percentiles = td->o.clat_percentiles;
 	if (td->o.overwrite_plist)
-		td->ts.percentile_list = td->o.percentile_list;
+		memcpy(td->ts.percentile_list, td->o.percentile_list, sizeof(td->o.percentile_list));
 	else
-		td->ts.percentile_list = NULL;
+		memcpy(td->ts.percentile_list, def_percentile_list, sizeof(def_percentile_list));
 
 	td->ts.clat_stat[0].min_val = td->ts.clat_stat[1].min_val = ULONG_MAX;
 	td->ts.slat_stat[0].min_val = td->ts.slat_stat[1].min_val = ULONG_MAX;
@@ -645,19 +829,23 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	td->groupid = groupid;
 	prev_group_jobs++;
 
-	if (init_random_state(td))
+	if (init_random_state(td, td->rand_seeds, sizeof(td->rand_seeds))) {
+		td_verror(td, errno, "init_random_state");
 		goto err;
+	}
 
 	if (setup_rate(td))
 		goto err;
 
 	if (td->o.write_lat_log) {
-		setup_log(&td->ts.lat_log);
-		setup_log(&td->ts.slat_log);
-		setup_log(&td->ts.clat_log);
+		setup_log(&td->lat_log, td->o.log_avg_msec);
+		setup_log(&td->slat_log, td->o.log_avg_msec);
+		setup_log(&td->clat_log, td->o.log_avg_msec);
 	}
 	if (td->o.write_bw_log)
-		setup_log(&td->ts.bw_log);
+		setup_log(&td->bw_log, td->o.log_avg_msec);
+	if (td->o.write_iops_log)
+		setup_log(&td->iops_log, td->o.log_avg_msec);
 
 	if (!td->o.name)
 		td->o.name = strdup(jobname);
@@ -700,7 +888,7 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 	 */
 	numjobs = td->o.numjobs;
 	while (--numjobs) {
-		struct thread_data *td_new = get_new_job(0, td);
+		struct thread_data *td_new = get_new_job(0, td, 1);
 
 		if (!td_new)
 			goto err;
@@ -748,11 +936,11 @@ void add_job_opts(const char **o)
 			sprintf(jobname, "%s", o[i] + 5);
 		}
 		if (in_global && !td_parent)
-			td_parent = get_new_job(1, &def_thread);
+			td_parent = get_new_job(1, &def_thread, 0);
 		else if (!in_global && !td) {
 			if (!td_parent)
 				td_parent = &def_thread;
-			td = get_new_job(0, td_parent);
+			td = get_new_job(0, td_parent, 0);
 		}
 		if (in_global)
 			fio_options_parse(td_parent, (char **) &o[i], 1);
@@ -800,7 +988,7 @@ static int is_empty_or_comment(char *line)
 /*
  * This is our [ini] type file parser.
  */
-static int parse_jobs_ini(char *file, int stonewall_flag)
+int parse_jobs_ini(char *file, int is_buf, int stonewall_flag)
 {
 	unsigned int global;
 	struct thread_data *td;
@@ -814,14 +1002,18 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 	char **opts;
 	int i, alloc_opts, num_opts;
 
-	if (!strcmp(file, "-"))
-		f = stdin;
-	else
-		f = fopen(file, "r");
+	if (is_buf)
+		f = NULL;
+	else {
+		if (!strcmp(file, "-"))
+			f = stdin;
+		else
+			f = fopen(file, "r");
 
-	if (!f) {
-		perror("fopen job file");
-		return 1;
+		if (!f) {
+			perror("fopen job file");
+			return 1;
+		}
 	}
 
 	string = malloc(4096);
@@ -843,7 +1035,10 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 		 * haven't handled.
 		 */
 		if (!skip_fgets) {
-			p = fgets(string, 4095, f);
+			if (is_buf)
+				p = strsep(&file, "\n");
+			else
+				p = fgets(string, 4096, f);
 			if (!p)
 				break;
 		}
@@ -880,7 +1075,7 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 			first_sect = 0;
 		}
 
-		td = get_new_job(global, &def_thread);
+		td = get_new_job(global, &def_thread, 0);
 		if (!td) {
 			ret = 1;
 			break;
@@ -897,7 +1092,14 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 		num_opts = 0;
 		memset(opts, 0, alloc_opts * sizeof(char *));
 
-		while ((p = fgets(string, 4096, f)) != NULL) {
+		while (1) {
+			if (is_buf)
+				p = strsep(&file, "\n");
+			else
+				p = fgets(string, 4096, f);
+			if (!p)
+				break;
+
 			if (is_empty_or_comment(p))
 				continue;
 
@@ -944,13 +1146,19 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 	if (dump_cmdline)
 		log_info("\n");
 
+	i = 0;
+	while (i < nr_job_sections) {
+		free(job_sections[i]);
+		i++;
+	}
+
 	for (i = 0; i < num_opts; i++)
 		free(opts[i]);
 
 	free(string);
 	free(name);
 	free(opts);
-	if (f != stdin)
+	if (!is_buf && f != stdin)
 		fclose(f);
 	return ret;
 }
@@ -969,99 +1177,42 @@ static int fill_def_thread(void)
 	return 0;
 }
 
-static void free_shm(void)
-{
-	struct shmid_ds sbuf;
-
-	if (threads) {
-		void *tp = threads;
-
-		threads = NULL;
-		file_hash_exit();
-		fio_debug_jobp = NULL;
-		shmdt(tp);
-		shmctl(shm_id, IPC_RMID, &sbuf);
-	}
-
-	scleanup();
-}
-
-/*
- * The thread area is shared between the main process and the job
- * threads/processes. So setup a shared memory segment that will hold
- * all the job info. We use the end of the region for keeping track of
- * open files across jobs, for file sharing.
- */
-static int setup_thread_area(void)
-{
-	void *hash;
-
-	/*
-	 * 1024 is too much on some machines, scale max_jobs if
-	 * we get a failure that looks like too large a shm segment
-	 */
-	do {
-		size_t size = max_jobs * sizeof(struct thread_data);
-
-		size += file_hash_size;
-		size += sizeof(unsigned int);
-
-		shm_id = shmget(0, size, IPC_CREAT | 0600);
-		if (shm_id != -1)
-			break;
-		if (errno != EINVAL) {
-			perror("shmget");
-			break;
-		}
-
-		max_jobs >>= 1;
-	} while (max_jobs);
-
-	if (shm_id == -1)
-		return 1;
-
-	threads = shmat(shm_id, NULL, 0);
-	if (threads == (void *) -1) {
-		perror("shmat");
-		return 1;
-	}
-
-	memset(threads, 0, max_jobs * sizeof(struct thread_data));
-	hash = (void *) threads + max_jobs * sizeof(struct thread_data);
-	fio_debug_jobp = (void *) hash + file_hash_size;
-	*fio_debug_jobp = -1;
-	file_hash_init(hash);
-	atexit(free_shm);
-	return 0;
-}
-
 static void usage(const char *name)
 {
 	printf("%s\n", fio_version_string);
 	printf("%s [options] [job options] <job file(s)>\n", name);
-	printf("\t--debug=options\tEnable debug logging\n");
-	printf("\t--output\tWrite output to file\n");
-	printf("\t--timeout\tRuntime in seconds\n");
-	printf("\t--latency-log\tGenerate per-job latency logs\n");
-	printf("\t--bandwidth-log\tGenerate per-job bandwidth logs\n");
-	printf("\t--minimal\tMinimal (terse) output\n");
-	printf("\t--version\tPrint version info and exit\n");
-	printf("\t--terse-version=x Terse version output format\n");
-	printf("\t--help\t\tPrint this page\n");
-	printf("\t--cmdhelp=cmd\tPrint command help, \"all\" for all of"
+	printf("  --debug=options\tEnable debug logging. May be one/more of:\n"
+		"\t\t\tprocess,file,io,mem,blktrace,verify,random,parse,\n"
+		"\t\t\tdiskutil,job,mutex,profile,time,net\n");
+	printf("  --output\t\tWrite output to file\n");
+	printf("  --timeout\t\tRuntime in seconds\n");
+	printf("  --latency-log\t\tGenerate per-job latency logs\n");
+	printf("  --bandwidth-log\tGenerate per-job bandwidth logs\n");
+	printf("  --minimal\t\tMinimal (terse) output\n");
+	printf("  --version\t\tPrint version info and exit\n");
+	printf("  --terse-version=x\tSet terse version output format to 'x'\n");
+	printf("  --help\t\tPrint this page\n");
+	printf("  --cmdhelp=cmd\t\tPrint command help, \"all\" for all of"
 		" them\n");
-	printf("\t--showcmd\tTurn a job file into command line options\n");
-	printf("\t--eta=when\tWhen ETA estimate should be printed\n");
-	printf("\t          \tMay be \"always\", \"never\" or \"auto\"\n");
-	printf("\t--readonly\tTurn on safety read-only checks, preventing"
+	printf("  --enghelp=engine\tPrint ioengine help, or list"
+		" available ioengines\n");
+	printf("  --enghelp=engine,cmd\tPrint help for an ioengine"
+		" cmd\n");
+	printf("  --showcmd\t\tTurn a job file into command line options\n");
+	printf("  --eta=when\t\tWhen ETA estimate should be printed\n");
+	printf("            \t\tMay be \"always\", \"never\" or \"auto\"\n");
+	printf("  --readonly\t\tTurn on safety read-only checks, preventing"
 		" writes\n");
-	printf("\t--section=name\tOnly run specified section in job file\n");
-	printf("\t--alloc-size=kb\tSet smalloc pool to this size in kb"
+	printf("  --section=name\tOnly run specified section in job file\n");
+	printf("  --alloc-size=kb\tSet smalloc pool to this size in kb"
 		" (def 1024)\n");
-	printf("\t--warnings-fatal Fio parser warnings are fatal\n");
-	printf("\t--max-jobs\tMaximum number of threads/processes to support\n");
+	printf("  --warnings-fatal\tFio parser warnings are fatal\n");
+	printf("  --max-jobs=nr\t\tMaximum number of threads/processes to support\n");
+	printf("  --server=args\t\tStart a backend fio server\n");
+	printf("  --daemonize=pidfile\tBackground fio server, write pid to file\n");
+	printf("  --client=hostname\tTalk to remote backend fio server at hostname\n");
 	printf("\nFio was written by Jens Axboe <jens.axboe@oracle.com>");
-	printf("\n                   Jens Axboe <jaxboe@fusionio.com>\n");
+	printf("\n                 Jens Axboe <jaxboe@fusionio.com>\n");
 }
 
 #ifdef FIO_INC_DEBUG
@@ -1079,6 +1230,7 @@ struct debug_level debug_levels[] = {
 	{ .name = "mutex",	.shift = FD_MUTEX },
 	{ .name	= "profile",	.shift = FD_PROFILE },
 	{ .name = "time",	.shift = FD_TIME },
+	{ .name = "net",	.shift = FD_NET },
 	{ .name = NULL, },
 };
 
@@ -1163,13 +1315,51 @@ static void fio_options_fill_optstring(void)
 	ostr[c] = '\0';
 }
 
-static int parse_cmd_line(int argc, char *argv[])
+static int client_flag_set(char c)
+{
+	int i;
+
+	i = 0;
+	while (l_opts[i].name) {
+		int val = l_opts[i].val;
+
+		if (c == (val & 0xff))
+			return (val & FIO_CLIENT_FLAG);
+
+		i++;
+	}
+
+	return 0;
+}
+
+void parse_cmd_client(void *client, char *opt)
+{
+	fio_client_add_cmd_option(client, opt);
+}
+
+int parse_cmd_line(int argc, char *argv[])
 {
 	struct thread_data *td = NULL;
 	int c, ini_idx = 0, lidx, ret = 0, do_exit = 0, exit_val = 0;
 	char *ostr = cmd_optstr;
+	void *pid_file = NULL;
+	void *cur_client = NULL;
+	int backend = 0;
+
+	/*
+	 * Reset optind handling, since we may call this multiple times
+	 * for the backend.
+	 */
+	optind = 1;
 
 	while ((c = getopt_long_only(argc, argv, ostr, l_opts, &lidx)) != -1) {
+		did_arg = 1;
+
+		if ((c & FIO_CLIENT_FLAG) || client_flag_set(c)) {
+			parse_cmd_client(cur_client, argv[optind - 1]);
+			c &= ~FIO_CLIENT_FLAG;
+		}
+
 		switch (c) {
 		case 'a':
 			smalloc_pool_size = atoi(optarg);
@@ -1195,10 +1385,23 @@ static int parse_cmd_line(int argc, char *argv[])
 			terse_output = 1;
 			break;
 		case 'h':
-			usage(argv[0]);
-			exit(0);
+			if (!cur_client) {
+				usage(argv[0]);
+				do_exit++;
+			}
+			break;
 		case 'c':
-			exit(fio_show_option_help(optarg));
+			if (!cur_client) {
+				fio_show_option_help(optarg);
+				do_exit++;
+			}
+			break;
+		case 'i':
+			if (!cur_client) {
+				fio_show_ioengine_help(optarg);
+				do_exit++;
+			}
+			break;
 		case 's':
 			dump_cmdline = 1;
 			break;
@@ -1206,11 +1409,14 @@ static int parse_cmd_line(int argc, char *argv[])
 			read_only = 1;
 			break;
 		case 'v':
-			log_info("%s\n", fio_version_string);
-			exit(0);
+			if (!cur_client) {
+				log_info("%s\n", fio_version_string);
+				do_exit++;
+			}
+			break;
 		case 'V':
 			terse_version = atoi(optarg);
-			if (terse_version != 2) {
+			if (!(terse_version == 2 || terse_version == 3)) {
 				log_err("fio: bad terse version format\n");
 				exit_val = 1;
 				do_exit++;
@@ -1265,12 +1471,28 @@ static int parse_cmd_line(int argc, char *argv[])
 				if (is_section && skip_this_section(val))
 					continue;
 
-				td = get_new_job(global, &def_thread);
-				if (!td)
+				td = get_new_job(global, &def_thread, 1);
+				if (!td || ioengine_load(td))
 					return 0;
+				fio_options_set_ioengine_opts(l_opts, td);
 			}
 
 			ret = fio_cmd_option_parse(td, opt, val);
+
+			if (!ret && !strcmp(opt, "ioengine")) {
+				free_ioengine(td);
+				if (ioengine_load(td))
+					return 0;
+				fio_options_set_ioengine_opts(l_opts, td);
+			}
+			break;
+		}
+		case FIO_GETOPT_IOENGINE: {
+			const char *opt = l_opts[lidx].name;
+			char *val = optarg;
+			opt = l_opts[lidx].name;
+			val = optarg;
+			ret = fio_cmd_ioengine_option_parse(td, opt, val);
 			break;
 		}
 		case 'w':
@@ -1284,22 +1506,64 @@ static int parse_cmd_line(int argc, char *argv[])
 				exit_val = 1;
 			}
 			break;
+		case 'S':
+			if (nr_clients) {
+				log_err("fio: can't be both client and server\n");
+				do_exit++;
+				exit_val = 1;
+				break;
+			}
+			if (optarg)
+				fio_server_set_arg(optarg);
+			is_backend = 1;
+			backend = 1;
+			break;
+		case 'D':
+			pid_file = strdup(optarg);
+			break;
+		case 'C':
+			if (is_backend) {
+				log_err("fio: can't be both client and server\n");
+				do_exit++;
+				exit_val = 1;
+				break;
+			}
+			if (fio_client_add(optarg, &cur_client)) {
+				log_err("fio: failed adding client %s\n", optarg);
+				do_exit++;
+				exit_val = 1;
+				break;
+			}
+			break;
 		default:
 			do_exit++;
 			exit_val = 1;
 			break;
 		}
+		if (do_exit)
+			break;
 	}
 
-	if (do_exit)
-		exit(exit_val);
+	if (do_exit) {
+		if (exit_val && !(is_backend || nr_clients))
+			exit(exit_val);
+	}
+
+	if (nr_clients && fio_clients_connect()) {
+		do_exit++;
+		exit_val = 1;
+		return -1;
+	}
+
+	if (is_backend && backend)
+		return fio_start_server(pid_file);
 
 	if (td) {
 		if (!ret)
 			ret = add_job(td, td->o.name ?: "fio", 0);
 	}
 
-	while (optind < argc) {
+	while (!ret && optind < argc) {
 		ini_idx++;
 		ini_file = realloc(ini_file, ini_idx * sizeof(char *));
 		ini_file[ini_idx - 1] = strdup(argv[optind]);
@@ -1319,33 +1583,50 @@ int parse_options(int argc, char *argv[])
 	fio_options_fill_optstring();
 	fio_options_dup_and_init(l_opts);
 
-	if (setup_thread_area())
-		return 1;
+	atexit(free_shm);
+
 	if (fill_def_thread())
 		return 1;
 
 	job_files = parse_cmd_line(argc, argv);
 
-	for (i = 0; i < job_files; i++) {
-		if (fill_def_thread())
-			return 1;
-		if (parse_jobs_ini(ini_file[i], i))
-			return 1;
-		free(ini_file[i]);
+	if (job_files > 0) {
+		for (i = 0; i < job_files; i++) {
+			if (fill_def_thread())
+				return 1;
+			if (nr_clients) {
+				if (fio_clients_send_ini(ini_file[i]))
+					return 1;
+				free(ini_file[i]);
+			} else if (!is_backend) {
+				if (parse_jobs_ini(ini_file[i], 0, i))
+					return 1;
+				free(ini_file[i]);
+			}
+		}
 	}
 
 	free(ini_file);
-	options_mem_free(&def_thread);
+	fio_options_free(&def_thread);
 
 	if (!thread_number) {
 		if (dump_cmdline)
 			return 0;
 		if (exec_profile)
 			return 0;
+		if (is_backend || nr_clients)
+			return 0;
+		if (did_arg)
+			return 0;
 
 		log_err("No jobs(s) defined\n\n");
-		usage(argv[0]);
-		return 1;
+
+		if (!did_arg) {
+			usage(argv[0]);
+			return 1;
+		}
+
+		return 0;
 	}
 
 	if (def_thread.o.gtod_offload) {
@@ -1354,6 +1635,8 @@ int parse_options(int argc, char *argv[])
 		fio_gtod_cpu = def_thread.o.gtod_cpu;
 	}
 
-	log_info("%s\n", fio_version_string);
+	if (!terse_output)
+		log_info("%s\n", fio_version_string);
+
 	return 0;
 }
